package org.eclipse.jdt.internal.compiler.classfmt;

package .tmp;

import java.io.IOException;
import java.io.InputStream;
import java.util.Map;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.env.ITypeAnnotationWalker;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;

public class ExternalAnnotationProvider {
  public static final String ANNOTATION_FILE_EXTENSION = "eea";
  
  public static final String CLASS_PREFIX = "class ";
  
  public static final String SUPER_PREFIX = "super ";
  
  public static final char NULLABLE = '0';
  
  public static final char NONNULL = '1';
  
  public static final char NO_ANNOTATION = '@';
  
  public static final String ANNOTATION_FILE_SUFFIX = ".eea";
  
  private static final String TYPE_PARAMETER_PREFIX = " <";
  
  private String typeName;
  
  String typeParametersAnnotationSource;
  
  Map<String, String> supertypeAnnotationSources;
  
  private Map<String, String> methodAnnotationSources;
  
  private Map<String, String> fieldAnnotationSources;
  
  SingleMarkerAnnotation NULLABLE_ANNOTATION;
  
  SingleMarkerAnnotation NONNULL_ANNOTATION;
  
  public ExternalAnnotationProvider(InputStream input, String typeName) throws IOException {
    this.typeName = typeName;
    initialize(input);
  }
  
  private void initialize(InputStream input) throws IOException {
    // Byte code:
    //   0: aconst_null
    //   1: astore_2
    //   2: aconst_null
    //   3: astore_3
    //   4: new java/io/LineNumberReader
    //   7: dup
    //   8: new java/io/InputStreamReader
    //   11: dup
    //   12: aload_1
    //   13: invokespecial <init> : (Ljava/io/InputStream;)V
    //   16: invokespecial <init> : (Ljava/io/Reader;)V
    //   19: astore #4
    //   21: aload #4
    //   23: invokevirtual readLine : ()Ljava/lang/String;
    //   26: aload_0
    //   27: getfield typeName : Ljava/lang/String;
    //   30: invokestatic assertClassHeader : (Ljava/lang/String;Ljava/lang/String;)V
    //   33: aload #4
    //   35: invokevirtual readLine : ()Ljava/lang/String;
    //   38: dup
    //   39: astore #5
    //   41: ifnonnull -> 55
    //   44: aload #4
    //   46: ifnull -> 54
    //   49: aload #4
    //   51: invokevirtual close : ()V
    //   54: return
    //   55: aload #5
    //   57: ldc ' <'
    //   59: invokevirtual startsWith : (Ljava/lang/String;)Z
    //   62: ifeq -> 133
    //   65: aload #4
    //   67: invokevirtual readLine : ()Ljava/lang/String;
    //   70: dup
    //   71: astore #5
    //   73: ifnonnull -> 87
    //   76: aload #4
    //   78: ifnull -> 86
    //   81: aload #4
    //   83: invokevirtual close : ()V
    //   86: return
    //   87: aload #5
    //   89: ldc ' <'
    //   91: invokevirtual startsWith : (Ljava/lang/String;)Z
    //   94: ifeq -> 133
    //   97: aload_0
    //   98: aload #5
    //   100: ldc ' <'
    //   102: invokevirtual length : ()I
    //   105: invokevirtual substring : (I)Ljava/lang/String;
    //   108: putfield typeParametersAnnotationSource : Ljava/lang/String;
    //   111: aload #4
    //   113: invokevirtual readLine : ()Ljava/lang/String;
    //   116: dup
    //   117: astore #5
    //   119: ifnonnull -> 133
    //   122: aload #4
    //   124: ifnull -> 132
    //   127: aload #4
    //   129: invokevirtual close : ()V
    //   132: return
    //   133: aconst_null
    //   134: astore #6
    //   136: aload #5
    //   138: invokevirtual trim : ()Ljava/lang/String;
    //   141: astore #5
    //   143: aload #5
    //   145: invokevirtual isEmpty : ()Z
    //   148: ifeq -> 154
    //   151: goto -> 536
    //   154: aconst_null
    //   155: astore #7
    //   157: aconst_null
    //   158: astore #8
    //   160: aload #5
    //   162: astore #9
    //   164: aload #9
    //   166: ldc 'super '
    //   168: invokevirtual startsWith : (Ljava/lang/String;)Z
    //   171: istore #10
    //   173: iload #10
    //   175: ifeq -> 190
    //   178: aload #9
    //   180: ldc 'super '
    //   182: invokevirtual length : ()I
    //   185: invokevirtual substring : (I)Ljava/lang/String;
    //   188: astore #9
    //   190: iconst_m1
    //   191: istore #11
    //   193: aload #4
    //   195: invokevirtual readLine : ()Ljava/lang/String;
    //   198: astore #5
    //   200: aload #5
    //   202: ifnull -> 235
    //   205: aload #5
    //   207: invokevirtual isEmpty : ()Z
    //   210: ifne -> 235
    //   213: aload #5
    //   215: iconst_0
    //   216: invokevirtual charAt : (I)C
    //   219: bipush #32
    //   221: if_icmpne -> 235
    //   224: aload #5
    //   226: iconst_1
    //   227: invokevirtual substring : (I)Ljava/lang/String;
    //   230: astore #7
    //   232: goto -> 242
    //   235: aload #4
    //   237: invokevirtual getLineNumber : ()I
    //   240: istore #11
    //   242: aload #4
    //   244: invokevirtual readLine : ()Ljava/lang/String;
    //   247: astore #5
    //   249: aload #5
    //   251: ifnull -> 536
    //   254: aload #5
    //   256: invokevirtual isEmpty : ()Z
    //   259: ifeq -> 265
    //   262: goto -> 536
    //   265: aload #5
    //   267: iconst_0
    //   268: invokevirtual charAt : (I)C
    //   271: bipush #32
    //   273: if_icmpeq -> 283
    //   276: aload #5
    //   278: astore #6
    //   280: goto -> 536
    //   283: aload #5
    //   285: iconst_1
    //   286: invokevirtual substring : (I)Ljava/lang/String;
    //   289: astore #8
    //   291: goto -> 295
    //   294: pop
    //   295: aload #7
    //   297: ifnull -> 305
    //   300: aload #8
    //   302: ifnonnull -> 343
    //   305: iload #11
    //   307: iconst_m1
    //   308: if_icmpne -> 318
    //   311: aload #4
    //   313: invokevirtual getLineNumber : ()I
    //   316: istore #11
    //   318: new java/io/IOException
    //   321: dup
    //   322: new java/lang/StringBuilder
    //   325: dup
    //   326: ldc 'Illegal format for annotation file at line '
    //   328: invokespecial <init> : (Ljava/lang/String;)V
    //   331: iload #11
    //   333: invokevirtual append : (I)Ljava/lang/StringBuilder;
    //   336: invokevirtual toString : ()Ljava/lang/String;
    //   339: invokespecial <init> : (Ljava/lang/String;)V
    //   342: athrow
    //   343: aload #8
    //   345: invokestatic trimTail : (Ljava/lang/String;)Ljava/lang/String;
    //   348: astore #8
    //   350: iload #10
    //   352: ifeq -> 418
    //   355: aload_0
    //   356: getfield supertypeAnnotationSources : Ljava/util/Map;
    //   359: ifnonnull -> 373
    //   362: aload_0
    //   363: new java/util/HashMap
    //   366: dup
    //   367: invokespecial <init> : ()V
    //   370: putfield supertypeAnnotationSources : Ljava/util/Map;
    //   373: aload_0
    //   374: getfield supertypeAnnotationSources : Ljava/util/Map;
    //   377: new java/lang/StringBuilder
    //   380: dup
    //   381: bipush #76
    //   383: invokestatic valueOf : (C)Ljava/lang/String;
    //   386: invokespecial <init> : (Ljava/lang/String;)V
    //   389: aload #9
    //   391: invokevirtual append : (Ljava/lang/String;)Ljava/lang/StringBuilder;
    //   394: aload #7
    //   396: invokevirtual append : (Ljava/lang/String;)Ljava/lang/StringBuilder;
    //   399: bipush #59
    //   401: invokevirtual append : (C)Ljava/lang/StringBuilder;
    //   404: invokevirtual toString : ()Ljava/lang/String;
    //   407: aload #8
    //   409: invokeinterface put : (Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;
    //   414: pop
    //   415: goto -> 536
    //   418: aload #7
    //   420: ldc '('
    //   422: invokevirtual contains : (Ljava/lang/CharSequence;)Z
    //   425: ifeq -> 481
    //   428: aload_0
    //   429: getfield methodAnnotationSources : Ljava/util/Map;
    //   432: ifnonnull -> 446
    //   435: aload_0
    //   436: new java/util/HashMap
    //   439: dup
    //   440: invokespecial <init> : ()V
    //   443: putfield methodAnnotationSources : Ljava/util/Map;
    //   446: aload_0
    //   447: getfield methodAnnotationSources : Ljava/util/Map;
    //   450: new java/lang/StringBuilder
    //   453: dup
    //   454: aload #9
    //   456: invokestatic valueOf : (Ljava/lang/Object;)Ljava/lang/String;
    //   459: invokespecial <init> : (Ljava/lang/String;)V
    //   462: aload #7
    //   464: invokevirtual append : (Ljava/lang/String;)Ljava/lang/StringBuilder;
    //   467: invokevirtual toString : ()Ljava/lang/String;
    //   470: aload #8
    //   472: invokeinterface put : (Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;
    //   477: pop
    //   478: goto -> 536
    //   481: aload_0
    //   482: getfield fieldAnnotationSources : Ljava/util/Map;
    //   485: ifnonnull -> 499
    //   488: aload_0
    //   489: new java/util/HashMap
    //   492: dup
    //   493: invokespecial <init> : ()V
    //   496: putfield fieldAnnotationSources : Ljava/util/Map;
    //   499: aload_0
    //   500: getfield fieldAnnotationSources : Ljava/util/Map;
    //   503: new java/lang/StringBuilder
    //   506: dup
    //   507: aload #9
    //   509: invokestatic valueOf : (Ljava/lang/Object;)Ljava/lang/String;
    //   512: invokespecial <init> : (Ljava/lang/String;)V
    //   515: bipush #58
    //   517: invokevirtual append : (C)Ljava/lang/StringBuilder;
    //   520: aload #7
    //   522: invokevirtual append : (Ljava/lang/String;)Ljava/lang/StringBuilder;
    //   525: invokevirtual toString : ()Ljava/lang/String;
    //   528: aload #8
    //   530: invokeinterface put : (Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;
    //   535: pop
    //   536: aload #6
    //   538: dup
    //   539: astore #5
    //   541: ifnonnull -> 133
    //   544: aload #4
    //   546: invokevirtual readLine : ()Ljava/lang/String;
    //   549: dup
    //   550: astore #5
    //   552: ifnonnull -> 133
    //   555: aload #4
    //   557: ifnull -> 603
    //   560: aload #4
    //   562: invokevirtual close : ()V
    //   565: goto -> 603
    //   568: astore_2
    //   569: aload #4
    //   571: ifnull -> 579
    //   574: aload #4
    //   576: invokevirtual close : ()V
    //   579: aload_2
    //   580: athrow
    //   581: astore_3
    //   582: aload_2
    //   583: ifnonnull -> 591
    //   586: aload_3
    //   587: astore_2
    //   588: goto -> 601
    //   591: aload_2
    //   592: aload_3
    //   593: if_acmpeq -> 601
    //   596: aload_2
    //   597: aload_3
    //   598: invokevirtual addSuppressed : (Ljava/lang/Throwable;)V
    //   601: aload_2
    //   602: athrow
    //   603: return
    // Line number table:
    //   Java source line number -> byte code offset
    //   #72	-> 0
    //   #73	-> 21
    //   #76	-> 33
    //   #139	-> 44
    //   #77	-> 54
    //   #79	-> 55
    //   #80	-> 65
    //   #139	-> 76
    //   #81	-> 86
    //   #82	-> 87
    //   #83	-> 97
    //   #84	-> 111
    //   #139	-> 122
    //   #85	-> 132
    //   #90	-> 133
    //   #91	-> 136
    //   #92	-> 143
    //   #93	-> 154
    //   #95	-> 160
    //   #96	-> 164
    //   #97	-> 173
    //   #98	-> 178
    //   #99	-> 190
    //   #102	-> 193
    //   #103	-> 200
    //   #104	-> 224
    //   #106	-> 235
    //   #108	-> 242
    //   #109	-> 249
    //   #110	-> 262
    //   #111	-> 265
    //   #112	-> 276
    //   #113	-> 280
    //   #115	-> 283
    //   #116	-> 291
    //   #119	-> 295
    //   #120	-> 305
    //   #121	-> 318
    //   #124	-> 343
    //   #125	-> 350
    //   #126	-> 355
    //   #127	-> 362
    //   #128	-> 373
    //   #129	-> 415
    //   #130	-> 428
    //   #131	-> 435
    //   #132	-> 446
    //   #133	-> 478
    //   #134	-> 481
    //   #135	-> 488
    //   #136	-> 499
    //   #138	-> 536
    //   #139	-> 555
    //   #140	-> 603
    // Local variable table:
    //   start	length	slot	name	descriptor
    //   0	604	0	this	Lorg/eclipse/jdt/internal/compiler/classfmt/ExternalAnnotationProvider;
    //   0	604	1	input	Ljava/io/InputStream;
    //   21	558	4	reader	Ljava/io/LineNumberReader;
    //   41	514	5	line	Ljava/lang/String;
    //   136	419	6	pendingLine	Ljava/lang/String;
    //   157	379	7	rawSig	Ljava/lang/String;
    //   160	376	8	annotSig	Ljava/lang/String;
    //   164	372	9	selector	Ljava/lang/String;
    //   173	363	10	isSuper	Z
    //   193	343	11	errLine	I
    // Exception table:
    //   from	to	target	type
    //   4	581	581	finally
    //   21	44	568	finally
    //   54	76	568	finally
    //   86	122	568	finally
    //   132	555	568	finally
    //   193	262	294	java/lang/Exception
    //   265	280	294	java/lang/Exception
    //   283	291	294	java/lang/Exception
  }
  
  public static void assertClassHeader(String line, String typeName) throws IOException {
    if (line != null && line.startsWith("class ")) {
      line = line.substring("class ".length());
    } else {
      throw new IOException("missing class header in annotation file");
    } 
    if (!trimTail(line).equals(typeName))
      throw new IOException("mismatching class name in annotation file, expected " + typeName + ", but header said " + line); 
  }
  
  public static String extractSignature(String line) {
    if (line == null || line.isEmpty() || line.charAt(0) != ' ')
      return null; 
    return trimTail(line.substring(1));
  }
  
  protected static String trimTail(String line) {
    int tail = line.indexOf(' ');
    if (tail == -1)
      tail = line.indexOf('\t'); 
    if (tail != -1)
      return line.substring(0, tail); 
    return line;
  }
  
  public ITypeAnnotationWalker forTypeHeader(LookupEnvironment environment) {
    if (this.typeParametersAnnotationSource != null || this.supertypeAnnotationSources != null)
      return (ITypeAnnotationWalker)new DispatchingAnnotationWalker(this, environment); 
    return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
  }
  
  public ITypeAnnotationWalker forMethod(char[] selector, char[] signature, LookupEnvironment environment) {
    Map<String, String> sources = this.methodAnnotationSources;
    if (sources != null) {
      String source = sources.get(String.valueOf(CharOperation.concat(selector, signature)));
      if (source != null)
        return (ITypeAnnotationWalker)new MethodAnnotationWalker(this, source.toCharArray(), 0, environment); 
    } 
    return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
  }
  
  public ITypeAnnotationWalker forField(char[] selector, char[] signature, LookupEnvironment environment) {
    if (this.fieldAnnotationSources != null) {
      String source = this.fieldAnnotationSources.get(String.valueOf(CharOperation.concat(selector, signature, ':')));
      if (source != null)
        return (ITypeAnnotationWalker)new FieldAnnotationWalker(this, source.toCharArray(), 0, environment); 
    } 
    return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
  }
  
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append("External Annotations for ").append(this.typeName).append('\n');
    sb.append("Methods:\n");
    if (this.methodAnnotationSources != null)
      for (Map.Entry<String, String> e : this.methodAnnotationSources.entrySet())
        sb.append('\t').append(e.getKey()).append('\n');  
    return sb.toString();
  }
  
  void initAnnotations(LookupEnvironment environment) {
    if (this.NULLABLE_ANNOTATION == null)
      this.NULLABLE_ANNOTATION = (SingleMarkerAnnotation)new Object(this, this, environment); 
    if (this.NONNULL_ANNOTATION == null)
      this.NONNULL_ANNOTATION = (SingleMarkerAnnotation)new Object(this, this, environment); 
  }
}
